@@ -6,6 +6,10 @@
 #include "ksh_stat.h"
 #include <ctype.h>
 
+#ifdef __CYGWIN__
+#include <sys/cygwin.h>
+#endif /* __CYGWIN__ */
+
 int
 c_cd(wp)
 	char	**wp;
@@ -123,9 +127,12 @@ c_cd(wp)
 	/* Clear out tracked aliases with relative paths */
 	flushcom(0);
 
-	/* Set OLDPWD */
+	/* Set OLDPWD (note: unsetting OLDPWD does not disable this
+	 * setting in at&t ksh)
+	 */
 	if (current_wd[0])
-		setstr(oldpwd_s, current_wd);
+		/* Ignore failure (happens if readonly or integer) */
+		setstr(oldpwd_s, current_wd, KSH_RETURN_ERROR);
 
 	if (!ISABSPATH(Xstring(xs, xp))) {
 #ifdef OS2
@@ -145,8 +152,15 @@ c_cd(wp)
 
 	/* Set PWD */
 	if (pwd) {
-		set_current_wd(pwd);
-		setstr(pwd_s, pwd);
+#ifdef __CYGWIN__
+		char ptmp[PATH];  /* larger than MAX_PATH */
+		cygwin_conv_to_full_posix_path(pwd, ptmp);
+#else /* __CYGWIN__ */
+		char *ptmp = pwd;
+#endif /* __CYGWIN__ */
+		set_current_wd(ptmp);
+		/* Ignore failure (happens if readonly or integer) */
+		setstr(pwd_s, ptmp, KSH_RETURN_ERROR);
 	} else {
 		set_current_wd(null);
 		pwd = Xstring(xs, xp);
@@ -599,7 +613,7 @@ c_typeset(wp)
 	 * to get a number that is used with -L, -R, -Z or -i (eg, -1R2
 	 * sets right justify in a field of 12).  This allows options
 	 * to be grouped in an order (eg, -Lu12), but disallows -i8 -L3 and
-	 * does not allow the number to be specified as a seperate argument
+	 * does not allow the number to be specified as a separate argument
 	 * Here, the number must follow the RLZi option, but is optional
 	 * (see the # kludge in ksh_getopt()).
 	 */
@@ -763,15 +777,37 @@ c_typeset(wp)
 	    }
 	} else {
 	    for (l = e->loc; l; l = l->next) {
-		for (p = tsort(&l->vars); (vp = *p++); )
+		for (p = tsort(&l->vars); (vp = *p++); ) {
+		    struct tbl *tvp;
+		    int any_set = 0;
+		    /*
+		     * See if the parameter is set (for arrays, if any
+		     * element is set).
+		     */
+		    for (tvp = vp; tvp; tvp = tvp->u.array)
+			if (tvp->flag & ISSET) {
+			    any_set = 1;
+			    break;
+			}
+		    /*
+		     * Check attributes - note that all array elements
+		     * have (should have?) the same attributes, so checking
+		     * the first is sufficient.
+		     *
+		     * Report an unset param only if the user has
+		     * explicitly given it some attribute (like export);
+		     * otherwise, after "echo $FOO", we would report FOO...
+		     */
+		    if (!any_set && !(vp->flag & USERATTRIB))
+			continue;
+		    if (flag && (vp->flag & flag) == 0)
+			continue;
 		    for (; vp; vp = vp->u.array) {
-			/* Report an unset param only if the user has
-			 * explicitly given it some attribute (like export);
-			 * otherwise, after "echo $FOO", we would report FOO...
+			/* Ignore array elements that aren't set unless there
+			 * are no set elements, in which case the first is
+			 * reported on
 			 */
-			if (!(vp->flag & ISSET) && !(vp->flag & USERATTRIB))
-			    continue;
-			if (flag && (vp->flag & flag) == 0)
+			if ((vp->flag&ARRAY) && any_set && !(vp->flag & ISSET))
 			    continue;
 			/* no arguments */
 			if (thing == 0 && flag == 0) {
@@ -800,15 +836,14 @@ c_typeset(wp)
 				shprintf("-u ");
 			    if ((vp->flag&INT_U)) 
 				shprintf("-U ");
+			    shprintf("%s\n", vp->name);
 			    if (vp->flag&ARRAY)
-				shprintf("%s[%d]\n", vp->name, vp->index);
-			    else
-				shprintf("%s\n", vp->name);
+				break;
 			} else {
 			    if (pflag)
 				shprintf("%s ",
 				    (flag & EXPORT) ?  "export" : "readonly");
-			    if (vp->flag&ARRAY)
+			    if ((vp->flag&ARRAY) && any_set)
 				shprintf("%s[%d]", vp->name, vp->index);
 			    else
 				shprintf("%s", vp->name);
@@ -825,7 +860,13 @@ c_typeset(wp)
 			    }
 			    shprintf(newline);
 			}
+			/* Only report first `element' of an array with
+			 * no set elements.
+			 */
+			if (!any_set)
+			    break;
 		    }
+		}
 	    }
 	}
 	return 0;
@@ -1033,7 +1074,7 @@ c_let(wp)
 		bi_errorf("no arguments");
 	else
 		for (wp++; *wp; wp++)
-			if (!evaluate(*wp, &val, TRUE)) {
+			if (!evaluate(*wp, &val, KSH_RETURN_ERROR)) {
 				rv = 2;	/* distinguish error from zero result */
 				break;
 			} else
@@ -1148,7 +1189,7 @@ c_kill(wp)
 
 	/* assume old style options if -digits or -UPPERCASE */
 	if ((p = wp[1]) && *p == '-' && (digit(p[1]) || isupper(p[1]))) {
-		if (!(t = gettrap(p + 1))) {
+		if (!(t = gettrap(p + 1, TRUE))) {
 			bi_errorf("bad signal `%s'", p + 1);
 			return 1;
 		}
@@ -1162,7 +1203,7 @@ c_kill(wp)
 				lflag = 1;
 				break;
 			  case 's':
-				if (!(t = gettrap(builtin_opt.optarg))) {
+				if (!(t = gettrap(builtin_opt.optarg, TRUE))) {
 					bi_errorf("bad signal `%s'",
 						builtin_opt.optarg);
 					return 1;
@@ -1265,8 +1306,9 @@ c_getopts(wp)
 	const char *options;
 	const char *var;
 	int	optc;
+	int	ret;
 	char	buf[3];
-	struct tbl *vq;
+	struct tbl *vq, *voptarg;
 
 	if (ksh_getopt(wp, &builtin_opt, null) == '?')
 		return 1;
@@ -1332,21 +1374,27 @@ c_getopts(wp)
 		user_opt.uoptind = user_opt.optind;
 	}
 
+	voptarg = global("OPTARG");
+	voptarg->flag &= ~RDONLY;	/* at&t ksh clears ro and int */
+	/* Paranoia: ensure no bizarre results. */
+	if (voptarg->flag & INTEGER)
+	    typeset("OPTARG", 0, INTEGER, 0, 0);
 	if (user_opt.optarg == (char *) 0)
-		unset(global("OPTARG"), 0);
+		unset(voptarg, 0);
 	else
-		setstr(global("OPTARG"), user_opt.optarg);
+		/* This can't fail (have cleared readonly/integer) */
+		setstr(voptarg, user_opt.optarg, KSH_RETURN_ERROR);
+
+	ret = 0;
 
 	vq = global(var);
-	if (vq->flag & RDONLY) {
-		bi_errorf("%s is readonly", var);
-		return 1;
-	}
+	/* Error message already printed (integer, readonly) */
+	if (!setstr(vq, buf, KSH_RETURN_ERROR))
+	    ret = 1;
 	if (Flag(FEXPORT))
 		typeset(var, EXPORT, 0, 0, 0);
-	setstr(vq, buf);
 
-	return optc < 0 ? 1 : 0;
+	return optc < 0 ? 1 : ret;
 }
 
 #ifdef EMACS
