@@ -39,9 +39,9 @@ static char vdisable_c;
 void
 x_init()
 {
-	/* set to -1 to force initial binding */
+	/* set to -2 to force initial binding */
 	edchars.erase = edchars.kill = edchars.intr = edchars.quit
-		= edchars.eof = -1;
+		= edchars.eof = -2;
 	/* default value for deficient systems */
 	edchars.werase = 027;	/* ^W */
 
@@ -284,14 +284,33 @@ x_mode(onoff)
 
 		set_tty(tty_fd, &cb, TF_WAIT);
 
+#ifdef __CYGWIN__
+		if (edchars.eof == '\0')
+			edchars.eof = '\4';
+#endif /* __CYGWIN__ */
+
+		/* Convert unset values to internal `unset' value */
+		if (edchars.erase == vdisable_c)
+			edchars.erase = -1;
+		if (edchars.kill == vdisable_c)
+			edchars.kill = -1;
+		if (edchars.intr == vdisable_c)
+			edchars.intr = -1;
+		if (edchars.quit == vdisable_c)
+			edchars.quit = -1;
+		if (edchars.eof == vdisable_c)
+			edchars.eof = -1;
+		if (edchars.werase == vdisable_c)
+			edchars.werase = -1;
 		if (memcmp(&edchars, &oldchars, sizeof(edchars)) != 0) {
 #ifdef EMACS
 			x_emacs_keys(&edchars);
 #endif
 		}
-	} else
+	} else {
 		/* TF_WAIT doesn't seem to be necessary when leaving xmode */
 		set_tty(tty_fd, &tty_state, TF_NONE);
+	}
 
 	return prev;
 }
@@ -703,7 +722,7 @@ x_command_glob(flags, str, slen, wordsp)
 	return nwords;
 }
 
-#define IS_WORDC(c)	!isspace(c)
+#define IS_WORDC(c)	!(ctype(c, C_LEX1) || (c) == '\'' || (c) == '"')
 
 static int
 x_locate_word(buf, buflen, pos, startp, is_commandp)
