@@ -263,7 +263,7 @@ expand(cp, wp, f)
 					v.type = 10; /* not default */
 					v.name[0] = '\0';
 					v_evaluate(&v, substitute(sp, 0),
-						FALSE);
+						KSH_UNWIND_ERROR);
 					sp = strchr(sp, 0) + 1;
 					for (p = str_val(&v); *p; ) {
 						Xcheck(ds, dp);
@@ -396,12 +396,6 @@ expand(cp, wp, f)
 					st = st->prev;
 					continue;
 				  case '=':
-					if (st->var->flag & RDONLY)
-						/* XXX POSIX says this is only
-						 * fatal for special builtins
-						 */
-						errorf("%s: is read only",
-							st->var->name);
 					/* Restore our position and substitute
 					 * the value of st->var (may not be
 					 * the assigned value in the presence
@@ -412,9 +406,17 @@ expand(cp, wp, f)
 					 * global would cause with things
 					 * like x[i+=1] to be evaluated twice.
 					 */
+					/* Note: not exported by FEXPORT
+					 * in at&t ksh.
+					 */
+					/* XXX POSIX says readonly is only
+					 * fatal for special builtins (setstr
+					 * does readonly check).
+					 */
 					setstr(st->var, debunk(
 						(char *) alloc(strlen(dp) + 1,
-							ATEMP), dp));
+							ATEMP), dp),
+						KSH_UNWIND_ERROR);
 					x.str = str_val(st->var);
 					type = XSUB;
 					if (f&DOBLANK)
@@ -711,13 +713,17 @@ varsub(xp, sp, word, stypep, slenp)
 		sp++;
 		/* Check for size of array */
 		if ((p=strchr(sp,'[')) && (p[1]=='*'||p[1]=='@') && p[2]==']') {
-			c = 0;
+			int n = 0;
+			int max = 0;
 			vp = global(arrayname(sp));
 			if (vp->flag & (ISSET|ARRAY))
 				zero_ok = 1;
 			for (; vp; vp = vp->u.array)
-				if (vp->flag&ISSET)
-					c = vp->index+1;
+				if (vp->flag & ISSET) {
+					max = vp->index + 1;
+					n++;
+				}
+			c = n; /* ksh88/ksh93 go for number, not max index */
 		} else if (c == '*' || c == '@')
 			c = e->loc->argc;
 		else {
