@@ -43,6 +43,7 @@ int 	c_unset		ARGS((char **wp));
 int 	c_ulimit	ARGS((char **wp));
 int 	c_times		ARGS((char **wp));
 int 	timex		ARGS((struct op *t, int f));
+void	timex_hook	ARGS((struct op *t, char ** volatile *app));
 int 	c_exec		ARGS((char **wp));
 int 	c_builtin	ARGS((char **wp));
 /* c_test.c */
@@ -88,7 +89,7 @@ int 	c_fc	 	ARGS((register char **wp));
 void	sethistsize	ARGS((int n));
 void	sethistfile	ARGS((const char *name));
 # ifdef EASY_HISTORY
-void 	histappend	ARGS((const char *cmd, int nl_seperate));
+void 	histappend	ARGS((const char *cmd, int nl_separate));
 # endif
 char **	histpos	 	ARGS((void));
 int 	histN	 	ARGS((void));
@@ -110,6 +111,12 @@ void 	shellf		ARGS((const char *fmt, ...))
 				GCC_FUNC_ATTR(format(printf, 1, 2));
 void 	shprintf	ARGS((const char *fmt, ...))
 				GCC_FUNC_ATTR(format(printf, 1, 2));
+#ifdef KSH_DEBUG
+void 	kshdebug_init_	ARGS((void));
+void 	kshdebug_printf_ ARGS((const char *fmt, ...))
+				GCC_FUNC_ATTR(format(printf, 1, 2));
+void 	kshdebug_dump_	ARGS((const char *str, const void *mem, int nbytes));
+#endif /* KSH_DEBUG */
 int	can_seek	ARGS((int fd));
 void	initio		ARGS((void));
 int	ksh_dup2	ARGS((int ofd, int nfd, int errok));
@@ -126,7 +133,7 @@ void	coproc_write_close ARGS((int fd));
 int	coproc_getfd	ARGS((int mode, const char **emsgp));
 void	coproc_cleanup	ARGS((int reuse));
 #endif /* KSH */
-struct temp *maketemp	ARGS((Area *ap));
+struct temp *maketemp	ARGS((Area *ap, Temp_type type, struct temp **tlist));
 /* jobs.c */
 void 	j_init		ARGS((int mflagset));
 void 	j_exit		ARGS((void));
@@ -151,6 +158,7 @@ void 	pprompt		ARGS((const char *cp, int ntruncate));
 /* mail.c */
 #ifdef KSH
 void 	mcheck		ARGS((void));
+void 	mcset		ARGS((long interval));
 void 	mbset		ARGS((char *p));
 void 	mpset		ARGS((char *mptoparse));
 #endif /* KSH */
@@ -220,7 +228,7 @@ void	inittraps	ARGS((void));
 #ifdef KSH
 void	alarm_init	ARGS((void));
 #endif /* KSH */
-Trap *	gettrap		ARGS((const char *name));
+Trap *	gettrap		ARGS((const char *name, int igncase));
 RETSIGTYPE trapsig	ARGS((int i));
 void	intrcheck	ARGS((void));
 int	fatal_trap_check ARGS((void));
@@ -250,7 +258,7 @@ struct tbl *	global	ARGS((const char *n));
 struct tbl *	local	ARGS((const char *n, bool_t copy));
 char *	str_val		ARGS((struct tbl *vp));
 long 	intval		ARGS((struct tbl *vp));
-void 	setstr		ARGS((struct tbl *vq, const char *s));
+int 	setstr		ARGS((struct tbl *vq, const char *s, int error_ok));
 struct tbl *setint_v	ARGS((struct tbl *vq, struct tbl *vp));
 void 	setint		ARGS((struct tbl *vq, long n));
 int	getint		ARGS((struct tbl *vp, long *nump));
@@ -261,6 +269,7 @@ char	*skip_wdvarname ARGS((const char *s, int aok));
 int	is_wdvarname	ARGS((const char *s, int aok));
 int	is_wdvarassign	ARGS((const char *s));
 char **	makenv		ARGS((void));
+void	change_random	ARGS((void));
 int	array_ref_len	ARGS((const char *cp));
 char *	arrayname	ARGS((const char *str));
 void    set_array	ARGS((const char *var, int reset, char **vals));
