@@ -5,4 +5,4 @@
 #include "sh.h"
 
 const char ksh_version [] =
-	"@(#)PD KSH v5.2.13 97/10/27";
+	"@(#)PD KSH v5.2.14 99/07/13.2";
